# !/usr/bin/env python
#
# A logging and statistics bot for Telegram based on python-telegram-bot.
# Copyright (C) 2020
# Michael DM Dryden <mk.dryden@utoronto.ca>
#
# This file is part of telegram-stats-bot.
#
# telegram-stats-bot is free software: you can redistribute it and/or modify
# it under the terms of the GNU Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Public License
# along with this program. If not, see [http://www.gnu.org/licenses/].

import logging
import json
import argparse
import shlex
import warnings
import os

import telegram
from telegram.error import BadRequest
from telegram.ext import Updater, CommandHandler, MessageHandler, Filters, CallbackContext, JobQueue
from telegram.update import Update
import appdirs

from .parse import parse_message
from .log_storage import JSONStore, PostgresStore
from .stats import StatsRunner, get_parser, HelpException


warnings.filterwarnings("ignore")

logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                    level=logging.INFO)
logger = logging.getLogger(__name__)

stats = None

try:
    with open("./sticker-keys.json", 'r') as f:
        stickers = json.load(f)
except FileNotFoundError:
    stickers = {}
sticker_idx = None
sticker_id = None


def log_message(update: Update, context: CallbackContext):
    if update.edited_message:
        edited_message, user = parse_message(update.effective_message)
        bak_store.append_data('edited-messages', edited_message)
        store.update_data('messages', edited_message)
        return

    try:
        logger.info(update.effective_message.message_id)
    except AttributeError:
        logger.warning("No effective_message attribute")
    message, user = parse_message(update.effective_message)

    if message:
        bak_store.append_data('messages', message)
        store.append_data('messages', message)
    if len(user) > 0:
        for i in user:
            if i:
                bak_store.append_data('user_events', i)
                store.append_data('user_events', i)


def get_chatid(update: Update, context: CallbackContext):
    context.bot.send_message(chat_id=update.effective_chat.id,
                             text=f"Chat id: {update.effective_chat.id}")


def test_can_read_all_group_messages(context: CallbackContext):
    if not context.bot.can_read_all_group_messages:
        logger.error("Bot privacy is set to enabled, cannot log messages!!!")


def update_usernames_wrapper(context: CallbackContext):
    context.dispatcher.run_async(update_usernames, context)


def update_usernames(context: CallbackContext):  # context.job.context contains the chat_id
    user_ids = stats._get_message_user_ids()
    db_users = stats._get_db_users()
    tg_users = {user_id: None for user_id in user_ids}
    to_update = {}
    for u_id in tg_users:
        try:
            user = context.bot.get_chat_member(chat_id=context.job.context, user_id=u_id).user
            tg_users[u_id] = user.name, user.full_name
            if tg_users[u_id] != db_users[u_id]:
                if tg_users[u_id][1] == db_users[u_id][1]:  # Flag these so we don't insert new row
                    to_update[u_id] = tg_users[u_id][0], None
                else:
                    to_update[u_id] = tg_users[u_id]
        except KeyError:  # First time user
            to_update[u_id] = tg_users[u_id]
        except BadRequest:  # Handle users no longer in chat or haven't messaged since bot joined
            logger.debug("Couldn't get user %s", u_id)  # debug level because will spam every hour
    stats._update_user_ids(to_update)
    if stats.users_lock.acquire(timeout=10):
        stats.users = stats._get_db_users()
        stats.users_lock.release()
    else:
        logger.warning("Couldn't acquire username lock.")
        return
    logger.info("Usernames updated")


def print_stats(update: Update, context: CallbackContext):
    if update.effective_user.id not in stats.users:
        return

    stats_parser = get_parser(stats)
    image = None

    try:
        ns = stats_parser.parse_args(shlex.split(" ".join(context.args)))
    except HelpException as e:
        text = e.msg
        send_help(text, context, update)
        return
    except argparse.ArgumentError as e:
        text = str(e)
        send_help(text, context, update)
        return
    else:
        args = vars(ns)
        func = args.pop('func')

        try:
            if args['user']:
                try:
                    uid = args['user']
                    args['user'] = uid, stats.users[uid][0]
                except KeyError:
                    send_help("unknown userid", context, update)
                    return
        except KeyError:
            pass

        try:
            if args['me'] and not args['user']:  # Lets auto-user work by ignoring auto-input me arg
                args['user'] = update.effective_user.id, update.effective_user.name
            del args['me']
        except KeyError:
            pass

        try:
            text, image = func(**args)
        except HelpException as e:
            text = e.msg
            send_help(text, context, update)
            return

    if text:
        context.bot.send_message(chat_id=update.effective_chat.id,
                                 text=text,
                                 parse_mode=telegram.ParseMode.MARKDOWN_V2)
    if image:
        context.bot.send_photo(chat_id=update.effective_chat.id, photo=image)


def send_help(text: str, context: CallbackContext, update: Update):
    """
    Send help text to user. Tries to send a direct message if possible.
    :param text: text to send
    :param context:
    :param update:
    :return:
    """
    try:
        context.bot.send_message(chat_id=update.effective_user.id,
                                 text=f"```\n{text}\n```",
                                 parse_mode=telegram.ParseMode.MARKDOWN_V2)
    except telegram.error.Unauthorized:  # If user has never chatted with bot
        context.bot.send_message(chat_id=update.effective_chat.id,
                                 text=f"```\n{text}\n```",
                                 parse_mode=telegram.ParseMode.MARKDOWN_V2)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('token', type=str, help="Telegram bot token")
    parser.add_argument('chat_id', type=int, help="Telegram chat id to monitor.")
    parser.add_argument('postgres_url', type=str, help="Sqlalchemy-compatible postgresql url.")
    parser.add_argument('--json-path', type=str,
                        help="Either full path to backup storage folder or prefix (will be stored in app data dir.",
                        default="chat")
    args = parser.parse_args()

    updater = Updater(token=args.token, use_context=True)
    dispatcher = updater.dispatcher

    path = args.json_path
    if not os.path.split(path)[0]:  # Empty string for left part of path
        path = os.path.join(appdirs.user_data_dir('telegram-stats-bot'), path)

    os.makedirs(path, exist_ok=True)
    bak_store = JSONStore(path)
    store = PostgresStore(args.postgres_url)
    stats = StatsRunner(store.engine)

    stats_handler = CommandHandler('stats', print_stats, filters=~Filters.update.edited_message, run_async=True)
    dispatcher.add_handler(stats_handler)

    chat_id_handler = CommandHandler('chatid', get_chatid, filters=~Filters.update.edited_message)
    dispatcher.add_handler(chat_id_handler)

    if args.chat_id != 0:
        log_handler = MessageHandler(Filters.chat(chat_id=args.chat_id), log_message)
        dispatcher.add_handler(log_handler)

    job_queue: JobQueue = updater.job_queue
    update_users_job = job_queue.run_repeating(update_usernames_wrapper, interval=3600, first=5, context=args.chat_id)
    test_privacy_job = job_queue.run_once(test_can_read_all_group_messages, 0)

    updater.start_polling()
    updater.idle()



# !/usr/bin/env python
#
# A logging and statistics bot for Telegram based on python-telegram-bot.
# Copyright (C) 2020
# Michael DM Dryden <mk.dryden@utoronto.ca>
#
# This file is part of telegram-stats-bot.
#
# telegram-stats-bot is free software: you can redistribute it and/or modify
# it under the terms of the GNU Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Public License
# along with this program. If not, see [http://www.gnu.org/licenses/].

import json
import typing

import pandas as pd

media_dict = {'sticker': 'sticker',
              'animation': 'animation',
              'video_file': 'video',
              'voice_message': 'voice',
              'audio_file': 'audio',
              'video_message': 'video_note'}

user_event_cat = pd.Categorical(['left', 'joined'])
message_type_cat = pd.Categorical(['migrate_from_group', 'text', 'pinned_message', 'photo', 'sticker',
                                   'new_chat_members', 'left_chat_member', 'animation', 'video',
                                   'location', 'new_chat_title', 'voice', 'audio',
                                   'new_chat_photo', 'video_note', 'poll'])


def text_list_parser(text: typing.Union[str, typing.Sequence]) -> str:
    if isinstance(text, str):
        return text
    out = ""
    for block in text:
        try:
            out += block['text']
        except TypeError:
            out += block
    return out


def convert_messages(df: pd.DataFrame) -> typing.Tuple[typing.List[dict], typing.List[dict]]:
    messages_out = []
    users_out = []
    for message in df.itertuples():
        message_dict = {'message_id': message.id,
                        'date': message.date,
                        'from_user': None,
                        'forward_from_message_id': None,
                        'forward_from': None,
                        'forward_from_chat': None,
                        'caption': "",
                        'text': "",
                        'sticker_set_name': "",
                        'new_chat_title': "",
                        'reply_to_message': None,
                        'file_id': None,
                        'type': None,
                        }
        user_event_dict = {}
        if message.type == 'message':
            if pd.notnull(message.from_id):
                message_dict['from_user'] = message.from_id

            if pd.notnull(message.forwarded_from):
                try:
                    message_dict['forward_from'] = int(message.forwarded_from)
                except ValueError:
                    pass

            if pd.notnull(message.reply_to_message_id):
                message_dict['reply_to_message'] = message.reply_to_message_id

            if pd.notnull(message.photo):
                message_dict['type'] = 'photo'
                if message.text != "":
                    message_dict['caption'] = text_list_parser(message.text)
            elif pd.notnull(message.media_type):
                if message.text != "":
                    message_dict['caption'] = text_list_parser(message.text)
                message_dict['type'] = media_dict[message.media_type]
                if message.media_type == 'sticker' and '.webp' not in message.file:
                    message_dict['file_id'] = message.file
            elif message.text != "":
                message_dict['type'] = 'text'
                message_dict['text'] = text_list_parser(message.text)
            elif pd.notnull(message.poll):
                message_dict['type'] = 'poll'
            elif pd.notnull(message.location_information):
                message_dict['type'] = 'location'

        elif message.type == 'service':
            if pd.notnull(message.actor_id):
                message_dict['from_user'] = message.actor_id

            if message.action == 'edit_group_title':
                message_dict['type'] = 'new_chat_title'
                message_dict['new_chat_title'] = message.title
            elif message.action == 'pin_message':
                message_dict['type'] = 'pinned_message'
            elif message.action == 'edit_group_photo':
                message_dict['type'] = 'new_chat_photo'
            elif message.action == 'invite_members' or message.action == 'join_group_by_link':
                message_dict['type'] = 'new_chat_members'
                try:
                    for i in message.members:
                        users_out.append({'message_id': message.id,
                                          'user_id': i,
                                          'date': message.date,
                                          'event': 'join'})
                except TypeError:
                    user_event_dict = {'message_id': message.id,
                                       'user_id': message.actor_id,
                                       'date': message.date,
                                       'event': 'join'}
            elif message.action == 'remove_members':
                message_dict['type'] = 'left_chat_member'
                for i in message.members:
                    users_out.append({'message_id': message.id,
                                      'user_id': i,
                                      'date': message.date,
                                      'event': 'left'})
            else:
                message_dict['type'] = message.action
        messages_out.append(message_dict)
        if user_event_dict != {}:
            users_out.append(user_event_dict)
    return messages_out, users_out


def parse_json(path: str):
    with open(path, encoding='utf-8') as f:
        js = json.load(f)
    chat = js['chats']['list'][1]['messages']
    df = pd.DataFrame(chat)



# !/usr/bin/env python
#
# A logging and statistics bot for Telegram based on python-telegram-bot.
# Copyright (C) 2020
# Michael DM Dryden <mk.dryden@utoronto.ca>
#
# This file is part of telegram-stats-bot.
#
# telegram-stats-bot is free software: you can redistribute it and/or modify
# it under the terms of the GNU Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Public License
# along with this program. If not, see [http://www.gnu.org/licenses/].

import logging
from typing import Dict, List, Tuple, Text, NoReturn, Union
from threading import Lock
from io import BytesIO
import argparse
import inspect
import re
from datetime import timedelta, datetime

import pandas as pd
import seaborn as sns
import numpy as np
from matplotlib.figure import Figure
from matplotlib.dates import date2num
from sqlalchemy.engine import Engine

from .utils import escape_markdown
from . import __version__

sns.set_context('paper')
sns.set_style('whitegrid')
sns.set_palette("Set2")

logger = logging.getLogger()


class HelpException(Exception):
    def __init__(self, msg: str = None):
        self.msg = msg


class InternalParser(argparse.ArgumentParser):
    def error(self, message: Text) -> NoReturn:
        try:
            raise  # Reraises mostly ArgumentError for bad arg
        except RuntimeError:
            raise HelpException(message)

    def print_help(self, file=None) -> None:
        raise HelpException(self.format_help())

    def _print_message(self, message: str, file=None) -> None:
        raise HelpException(message)

    def exit(self, status=None, message=None):
        pass


class StatsRunner(object):
    allowed_methods = {'counts': "get_chat_counts",
                       'hours': "get_counts_by_hour",
                       'days': "get_counts_by_day",
                       'week': "get_week_by_hourday",
                       'history': "get_message_history",
                       'titles': 'get_title_history',
                       'corr': "get_user_correlation",
                       'delta': "get_message_deltas",
                       'types': "get_type_stats",
                       'random': "get_random_message"}

    def __init__(self, engine: Engine, tz: str = 'America/Toronto'):
        self.engine = engine
        self.tz = tz

        self.users: Dict[int, Tuple[str, str]] = self._get_db_users()
        self.users_lock = Lock()

    def _get_message_user_ids(self) -> List[int]:
        """Returns list of unique user ids from messages in database."""
        with self.engine.connect() as con:
            result = con.execute("SELECT DISTINCT from_user FROM messages_utc;")
        return [user for user, in result.fetchall()]

    def _get_db_users(self) -> Dict[int, Tuple[str, str]]:
        """Returns dictionary mapping user ids to usernames and full names."""
        query = """
        select user_id, username, display_name from (
            select
                *,
                row_number() over(partition by user_id order by date desc) as rn
            from
                user_names
            ) t
        where t.rn = 1
        """

        with self.engine.connect() as con:
            result = con.execute(query)
        result = result.fetchall()

        return {user_id: (username, name) for user_id, username, name in result}

    def _update_user_ids(self, user_dict: Dict[int, Tuple[str, str]]):
        """
        Updates user names table with user_dict
        :param user_dict: mapping of user ids to (username, display name)
        """
        for uid in user_dict:
            username, display_name = user_dict[uid]
            sql_dict = {'uid': uid, 'username': username, 'display_name': display_name}
            query = """
            UPDATE user_names
            SET username = %(username)s
            WHERE user_id = %(uid)s AND username IS DISTINCT FROM %(username)s;
            """
            if display_name:
                query += """\n
                         INSERT INTO user_names(user_id, date, username, display_name)
                             VALUES (%(uid)s, current_timestamp, %(username)s, %(display_name)s);
                         """

            with self.engine.connect() as con:
                con.execute(query, sql_dict)

    def get_chat_counts(self, n: int = 20, lquery: str = None, start: str = None, end: str = None) -> Tuple[str, None]:
        """
        Get top chat users
        :param lquery: Limit results to lexical query (&, |, !, <n>)
        :param n: Number of users to show
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        """
        sql_dict = {}
        query_conditions = []

        if n <= 0:
            raise HelpException(f'n must be greater than 0, got: {n}')

        if lquery:
            query_conditions.append(f"text_index_col @@ to_tsquery('{lquery}')")

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        query_where = ""
        if query_conditions:
            query_where = f"WHERE {' AND '.join(query_conditions)}"

        query = f"""
                SELECT "from_user", COUNT(*)
                FROM "messages_utc"
                {query_where}
                GROUP BY "from_user"
                ORDER BY "count" DESC;
                """
        with self.engine.connect() as con:
            df = pd.read_sql_query(query, con, params=sql_dict, index_col='from_user')

        user_df = pd.Series(self.users, name="user")
        user_df = user_df.apply(lambda x: x[0])  # Take only @usernames
        df = df.join(user_df)
        df['Percent'] = df['count'] / df['count'].sum() * 100
        df = df[['user', 'count', 'Percent']]
        if lquery:
            df.columns = ['User', lquery, 'Percent']
        else:
            df.columns = ['User', 'Total Messages', 'Percent']
        df['User'] = df['User'].str.replace(r'[^\x00-\x7F]', "", regex=True)  # Drop emoji

        text = df.iloc[:n].to_string(index=False, header=True, float_format=lambda x: f"{x:.1f}")

        return f"```\n{text}\n```", None

    def get_counts_by_hour(self, user: Tuple[int, str] = None, lquery: str = None, start: str = None, end: str = None) \
            -> Tuple[Union[str, None], Union[None, BytesIO]]:
        """
        Get plot of messages for hours of the day
        :param lquery: Limit results to lexical query (&, |, !, <n>)
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        """
        query_conditions = []
        sql_dict = {}

        if lquery:
            query_conditions.append(f"text_index_col @@ to_tsquery('{lquery}')")

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        if user:
            sql_dict['user'] = user[0]
            query_conditions.append("from_user = %(user)s")

        query_where = ""
        if query_conditions:
            query_where = f"WHERE {' AND '.join(query_conditions)}"

        query = f"""
                 SELECT date_trunc('hour', date) as day, count(*) as messages
                 FROM messages_utc
                 {query_where}
                 GROUP BY day
                 ORDER BY day
                 """

        with self.engine.connect() as con:
            df = pd.read_sql_query(query, con, params=sql_dict)

        if len(df) == 0:
            return "No matching messages", None

        df['day'] = pd.to_datetime(df.day)
        df['day'] = df.day.dt.tz_convert(self.tz)
        df = df.set_index('day')
        df = df.asfreq('h', fill_value=0)  # Insert 0s for periods with no messages

        if (df.index.max() - df.index.min()) < pd.Timedelta('24 hours'):  # Deal with data covering < 24 hours
            df = df.reindex(pd.date_range(df.index.min(), periods=24, freq='h'))

        df['hour'] = df.index.hour

        if user:
            # Aggregate over 1 week periods
            df = df.groupby('hour').resample('7D').sum().drop(columns='hour')
            df['hour'] = df.index.get_level_values('hour')

        fig = Figure(constrained_layout=True)
        subplot = fig.subplots()

        sns.stripplot(x='hour', y='messages', data=df, jitter=.4, size=2, ax=subplot, alpha=.5, zorder=0)
        sns.boxplot(x='hour', y='messages', data=df, whis=1, showfliers=False, whiskerprops={"zorder": 10},
                    boxprops={"zorder": 10},
                    ax=subplot, zorder=10)
        subplot.set_ylim(bottom=0, top=df['messages'].quantile(0.999, interpolation='higher'))

        subplot.axvspan(11.5, 23.5, zorder=0, color=(0, 0, 0, 0.05))
        subplot.set_xlim(-1, 24)  # Set explicitly to plot properly even with missing data

        if lquery:
            subplot.set_title(f"Messages by Hour for {lquery}")
        elif user:
            subplot.set_title(f"Messages by Hour for {user[1]}")
        if user:
            subplot.set_ylabel('Messages per Week')
        else:
            subplot.set_ylabel('Messages per Day')
            subplot.set_title("Messages by Hour")

        sns.despine(fig=fig)

        bio = BytesIO()
        bio.name = 'plot.png'
        fig.savefig(bio, bbox_inches='tight')
        bio.seek(0)

        return None, bio

    def get_counts_by_day(self, user: Tuple[int, str] = None, lquery: str = None, start: str = None, end: str = None,
                          plot: str = None) -> Tuple[Union[str, None], Union[None, BytesIO]]:
        """
        Get plot of messages for days of the week
        :param lquery: Limit results to lexical query (&, |, !, <n>)
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param plot: Type of plot. ('box' or 'violin')
        """
        query_conditions = []
        sql_dict = {}

        if lquery:
            query_conditions.append(f"text_index_col @@ to_tsquery('{lquery}')")

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        if user:
            sql_dict['user'] = user[0]
            query_conditions.append("from_user = %(user)s")

        query_where = ""
        if query_conditions:
            query_where = f"WHERE {' AND '.join(query_conditions)}"

        query = f"""
                 SELECT date_trunc('day', date)
                     as day, count(*) as messages
                 FROM messages_utc
                 {query_where}
                 GROUP BY day
                 ORDER BY day
                 """

        with self.engine.connect() as con:
            df = pd.read_sql_query(query, con, params=sql_dict)

        if len(df) == 0:
            return "No matching messages", None

        df['day'] = pd.to_datetime(df.day)
        df['day'] = df.day.dt.tz_convert(self.tz)
        df = df.set_index('day')
        df = df.asfreq('d', fill_value=0)  # Fill periods with no messages
        if (df.index.max() - df.index.min()) < pd.Timedelta('7 days'):  # Deal with data covering < 7 days
            df = df.reindex(pd.date_range(df.index.min(), periods=7, freq='d'))
        df['dow'] = df.index.weekday
        df['day_name'] = df.index.day_name()
        df = df.sort_values('dow')  # Make sure start is Monday

        fig = Figure(constrained_layout=True)
        subplot = fig.subplots()
        if plot == 'box':
            sns.boxplot(x='day_name', y='messages', data=df, whis=1, showfliers=False, ax=subplot)
        elif plot == 'violin' or plot is None:
            sns.violinplot(x='day_name', y='messages', data=df, cut=0, inner="box", scale='width', ax=subplot)
        else:
            raise HelpException("plot must be either box or violin")
        subplot.axvspan(4.5, 6.5, zorder=0, color=(0, .8, 0, 0.1))
        subplot.set_xlabel('')
        subplot.set_ylabel('Messages per Day')
        subplot.set_xlim(-0.5, 6.5)  # Need to set this explicitly to show full range of days with na data

        if lquery:
            subplot.set_title(f"Messages by Day of Week for {lquery}")
        elif user:
            subplot.set_title(f"Messages by Day of Week for {user[1]}")
        else:
            subplot.set_title("Messages by Day of Week")

        sns.despine(fig=fig)

        bio = BytesIO()
        bio.name = 'plot.png'
        fig.savefig(bio, bbox_inches='tight')
        bio.seek(0)

        return None, bio

    def get_week_by_hourday(self, lquery: str = None, user: Tuple[int, str] = None, start: str = None, end: str = None) \
            -> Tuple[Union[str, None], Union[None, BytesIO]]:
        """
        Get plot of messages over the week by day and hour.
        :param lquery: Limit results to lexical query (&, |, !, <n>)
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        """
        query_conditions = []
        sql_dict = {}

        if lquery:
            query_conditions.append(f"text_index_col @@ to_tsquery('{lquery}')")

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        if user:
            sql_dict['user'] = user[0]
            query_conditions.append("from_user = %(user)s")

        query_where = ""
        if query_conditions:
            query_where = f"WHERE {' AND '.join(query_conditions)}"

        query = f"""
                 SELECT date_trunc('hour', date)
                     as msg_time, count(*) as messages
                 FROM messages_utc
                 {query_where}
                 GROUP BY msg_time
                 ORDER BY msg_time
                 """
        with self.engine.connect() as con:
            df = pd.read_sql_query(query, con, params=sql_dict)

        if len(df) == 0:
            return "No matching messages", None

        df['msg_time'] = pd.to_datetime(df.msg_time)
        df['msg_time'] = df.msg_time.dt.tz_convert(self.tz)
        df = df.set_index('msg_time')
        df = df.asfreq('h', fill_value=0)  # Fill periods with no messages
        df['dow'] = df.index.weekday
        df['hour'] = df.index.hour
        df['day_name'] = df.index.day_name()
        df_grouped = df[['messages', 'hour', 'day_name']].groupby(['hour', 'day_name']).sum().unstack()
        df_grouped = df_grouped.loc[:, 'messages']
        df_grouped = df_grouped.reindex(columns=['Monday', 'Tuesday', 'Wednesday', 'Thursday',
                                                 'Friday', 'Saturday', 'Sunday'])

        fig = Figure(constrained_layout=True)
        ax = fig.subplots()

        sns.heatmap(df_grouped.T, yticklabels=['M', 'T', 'W', 'Th', 'F', 'Sa', 'Su'], linewidths=.5,
                    square=True, fmt='d', vmin=0,
                    cbar_kws={"orientation": "horizontal"}, cmap="viridis", ax=ax)
        ax.tick_params(axis='y', rotation=0)
        ax.set_ylabel("")
        ax.set_xlabel("")
        if lquery:
            ax.set_title(f"Messages by day and hour for {lquery}")
        elif user:
            ax.set_title(f"Total messages by day and hour for {user[1]}")
        else:
            ax.set_title("Total messages by day and hour")

        bio = BytesIO()
        bio.name = 'plot.png'
        fig.savefig(bio, bbox_inches='tight')
        bio.seek(0)

        return None, bio

    def get_message_history(self, user: Tuple[int, str] = None, lquery: str = None, averages: int = None, start: str = None,
                            end: str = None) \
            -> Tuple[Union[str, None], Union[None, BytesIO]]:
        """
        Make a plot of message history over time
        :param lquery: Limit results to lexical query (&, |, !, <n>)
        :param averages: Moving average width (in days)
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        """
        query_conditions = []
        sql_dict = {}

        if averages:
            if averages < 0:
                raise HelpException("averages must be >= 0")

        if lquery:
            query_conditions.append(f"text_index_col @@ to_tsquery('{lquery}')")

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        if user:
            sql_dict['user'] = user[0]
            query_conditions.append("from_user = %(user)s")

        query_where = ""
        if query_conditions:
            query_where = f"WHERE {' AND '.join(query_conditions)}"

        query = f"""
                    SELECT date_trunc('day', date)
                        as day, count(*) as messages
                    FROM messages_utc
                    {query_where}
                    GROUP BY day
                    ORDER BY day
                 """

        with self.engine.connect() as con:
            df = pd.read_sql_query(query, con, params=sql_dict)

        if len(df) == 0:
            return "No matching messages", None

        df['day'] = pd.to_datetime(df.day)
        df['day'] = df.day.dt.tz_convert(self.tz)

        if averages is None:
            averages = len(df) // 20
            if averages <= 1:
                averages = 0
        if averages:
            df['msg_rolling'] = df['messages'].rolling(averages, center=True).mean()
            alpha = 0.5
        else:
            alpha = 1

        fig = Figure(constrained_layout=True)
        subplot = fig.subplots()
        df.plot(x='day', y='messages', alpha=alpha, legend=False, ax=subplot)
        if averages:
            df.plot(x='day', y='msg_rolling', legend=False, ax=subplot)
        subplot.set_ylabel("Messages")
        subplot.set_xlabel("Date")
        if lquery:
            subplot.set_title(f"History for query: {lquery}")
        elif user:
            subplot.set_title(f"Message History for {user[1]}")
        else:
            subplot.set_title("Message History")
        sns.despine(fig=fig)
        fig.tight_layout()

        bio = BytesIO()
        bio.name = 'plot.png'
        fig.savefig(bio)
        bio.seek(0)

        return None, bio

    def get_title_history(self, start: str = None, end: str = None, duration: bool = False) \
            -> Tuple[Union[str, None], Union[None, BytesIO]]:
        """
        Make a plot of group titles history over time
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param duration: If true, order by duration instead of time.
        """
        query_conditions = []
        sql_dict = {}

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        query_where = ""
        if query_conditions:
            query_where = f"AND {' AND '.join(query_conditions)}"

        query = f"""
                    SELECT date, new_chat_title
                    FROM messages_utc
                    WHERE type = 'new_chat_title' {query_where}
                    ORDER BY date;
                 """

        with self.engine.connect() as con:
            df = pd.read_sql_query(query, con, params=sql_dict)

        df['idx'] = np.arange(len(df))
        df['diff'] = -df['date'].diff(-1)
        df['end'] = df['date'] + df['diff']

        if end:
            last = pd.Timestamp(sql_dict['end_dt'], tz=self.tz).tz_convert('utc')
        else:
            last = pd.Timestamp(datetime.now(), tz='utc')

        df_end = df['end']
        df_end.iloc[-1] = last
        df.loc[:, 'end'] = df_end
        df.loc[:, 'diff'].iloc[-1] = df.iloc[-1]['end'] - df.iloc[-1]['date']

        fig = Figure(constrained_layout=True, figsize=(12, 0.15 * len(df)))
        ax = fig.subplots()

        if duration:
            df = df.sort_values('diff')
            df = df.reset_index(drop=True)
            df['idx'] = df.index

            ax.barh(df.idx, df['diff'].dt.days, tick_label=df.new_chat_title)

            ax.margins(0.2)
            ax.set_ylabel("")
            ax.set_xlabel("Duration (days)")
            ax.set_ylim(-1, (df.idx.max() + 1))
            ax.set_title("Chat Title History")
            ax.grid(False, which='both', axis='y')
            sns.despine(fig=fig, left=True)

        else:
            x = df.iloc[:-1].end
            y = df.iloc[:-1].idx + .5

            ax.scatter(x, y, zorder=4, color=sns.color_palette()[1])

            titles = list(zip(df.date.apply(date2num),
                              df.end.apply(date2num) - df.date.apply(date2num)))

            for n, i in enumerate(titles):
                ax.broken_barh([i], (n, 1))
                ax.annotate(df.new_chat_title[n], xy=(i[0] + i[1], n), xycoords='data',
                            xytext=(10, 0), textcoords='offset points',
                            horizontalalignment='left', verticalalignment='bottom')

            ax.set_ylim(-1, (df.idx.max() + 1))
            ax.set_xlim(titles[0][0] - 1, None)

            ax.margins(0.2)
            ax.set_ylabel("")
            ax.set_xlabel("")
            ax.set_title("Chat Title History")
            ax.grid(False, which='both', axis='y')
            ax.tick_params(axis='y', which='both', labelleft=False, left=False)
            sns.despine(fig=fig, left=True)

        bio = BytesIO()
        bio.name = 'plot.png'
        fig.savefig(bio, dpi=200)
        bio.seek(0)

        return None, bio

    def get_user_correlation(self, start: str = None, end: str = None, agg: bool = True, c_type: str = None,
                             n: int = 5, thresh: float = 0.05, autouser=None, **kwargs) -> Tuple[str, None]:
        """
        Return correlations between you and other users.
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param agg: If True, calculate correlation over messages aggregated by hours of the week
        :param c_type: Correlation type to use. Either 'pearson' or 'spearman'
        :param n: Show n highest and lowest correlation scores
        :param thresh: Fraction of time bins that have data for both users to be considered valid (0-1)
        """
        user: Tuple[int, str] = kwargs['user']
        query_conditions = []
        sql_dict = {}

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        query_where = ""
        if query_conditions:
            query_where = f"WHERE {' AND '.join(query_conditions)}"

        if n <= 0:
            raise HelpException(f'n must be greater than 0, got: {n}')

        if not c_type:
            c_type = 'pearson'
        elif c_type not in ['pearson', 'spearman']:
            raise HelpException("corr must be either pearson or spearman")

        if not 0 <= thresh <= 1:
            raise HelpException(f'n must be in the range [0, 1], got: {n}')

        query = f"""
                SELECT msg_time, extract(ISODOW FROM msg_time) as dow, extract(HOUR FROM msg_time) as hour,
                       "user", messages
                FROM (
                         SELECT date_trunc('hour', date)
                                         as msg_time,
                                count(*) as messages, from_user as "user"
                         FROM messages_utc
                         {query_where}
                         GROUP BY msg_time, from_user
                         ORDER BY msg_time
                     ) t
                ORDER BY dow, hour;
                """

        with self.engine.connect() as con:
            df = pd.read_sql_query(query, con, params=sql_dict)
        df['msg_time'] = pd.to_datetime(df.msg_time)
        df['msg_time'] = df.msg_time.dt.tz_convert(self.tz)

        # Prune irrelevant messages (not sure if this actually improves performance)
        user_first_date = df.loc[df.user == user[0], 'msg_time'].iloc[0]
        df = df.loc[df.msg_time >= user_first_date]

        df = df.set_index('msg_time')

        user_dict = {'user': {user_id: value[0] for user_id, value in self.users.items()}}
        df = df.loc[df.user.isin(list(user_dict['user'].keys()))]  # Filter out users with no names
        df = df.replace(user_dict)  # Replace user ids with names
        df['user'] = df['user'].str.replace(r'[^\x00-\x7F]', "", regex=True)

        if agg:
            df = df.pivot_table(index=['dow', 'hour'], columns='user', values='messages', aggfunc='sum')
            corrs = []
            for other_user in df.columns.values:
                if df[user[1]].sum() / df[other_user].sum() > thresh:
                    me_notna = df[user[1]].notna()
                    other_notna = df[other_user].notna()
                    idx = me_notna | other_notna
                    corrs.append(df.loc[idx, user[1]].fillna(0).corr(df.loc[idx, other_user].fillna(0)))
                else:
                    corrs.append(pd.NA)

            me = pd.Series(corrs, index=df.columns.values).sort_values(ascending=False).iloc[1:].dropna()
        else:
            df = df.pivot(columns='user', values='messages')

            if thresh == 0:
                df_corr = df.corr(method=c_type)
            else:
                df_corr = df.corr(method=c_type, min_periods=int(thresh * len(df)))
            me = df_corr[user[1]].sort_values(ascending=False).iloc[1:].dropna()

        if len(me) < 1:
            return "`Sorry, not enough data, try with -aggtimes, decrease -thresh, or use a bigger date range.`", None

        if n > len(me) // 2:
            n = int(len(me) // 2)

        text = me.to_string(header=False, float_format=lambda x: f"{x:.3f}")
        split = text.splitlines()
        text = "\n".join(['HIGHEST CORRELATION:'] + split[:n] + ['\nLOWEST CORRELATION:'] + split[-n:])

        return f"**User Correlations for {escape_markdown(user[1])}**\n```\n{text}\n```", None

    def get_message_deltas(self, lquery: str = None, start: str = None, end: str = None, n: int = 10, thresh: int = 500,
                           autouser=None, **kwargs) -> Tuple[Union[str, None], Union[None, BytesIO]]:
        """
        Return the median difference in message time between you and other users.
        :param lquery: Limit results to lexical query (&, |, !, <n>)
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param n: Show n highest and lowest correlation scores
        :param thresh: Only consider users with at least this many message group pairs with you
        """
        user: Tuple[int, str] = kwargs['user']
        query_conditions = []
        sql_dict = {}

        if lquery:
            query_conditions.append(f"text_index_col @@ to_tsquery('{lquery}')")

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        query_where = ""
        if query_conditions:
            query_where = f"AND {' AND '.join(query_conditions)}"

        if n <= 0:
            raise HelpException(f'n must be greater than 0')

        if thresh < 0:
            raise HelpException(f'n cannot be negative')

        def fetch_mean_delta(me: int, other: int, where: str, sql_dict: dict) -> Tuple[timedelta, int]:
            query = f"""
                    select percentile_cont(0.5) within group (order by t_delta), count(t_delta)
                    from(
                        select start - lag("end", 1) over (order by start) as t_delta
                        from (
                                 select min(date) as start, max(date) as "end"
                                 from (select date, from_user,
                                              (dense_rank() over (order by date) -
                                               dense_rank() over (partition by from_user order by date)
                                                  ) as grp
                                       from messages_utc
                                       where from_user in (%(me)s, %(other)s) {where}
                                       order by date
                                      ) t
                                 group by from_user, grp
                                 order by start
                        ) t1
                    ) t2;
                    """

            sql_dict['me'] = me
            sql_dict['other'] = other

            with self.engine.connect() as con:
                result = con.execute(query, sql_dict)
            output: Tuple[timedelta, int] = result.fetchall()[0]

            return output

        results = {other: fetch_mean_delta(user[0], other, query_where, sql_dict) for other in self.users
                   if user[0] != other}

        user_deltas = {self.users[other][0]: pd.to_timedelta(result[0]) for other, result in results.items()
                       if result[1] > thresh}

        me = pd.Series(user_deltas).sort_values()
        me = me.apply(lambda x: x.round('1s'))

        if len(me) < 1:
            return "\n```\nSorry, not enough data, try a bigger date range or decrease -thresh.\n```", None

        text = me.iloc[:n].to_string(header=False, index=True)

        return f"**Median message delays for {escape_markdown(user[1])} and:**\n```\n{text}\n```", None

    def get_type_stats(self, start: str = None, end: str = None, autouser=None, **kwargs) -> Tuple[str, None]:
        """
        Print table of message statistics by type.
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        """
        user: Tuple[int, str] = kwargs['user']
        query_conditions = []
        sql_dict = {}

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        query_where = ""
        if query_conditions:
            query_where = f" AND {' AND '.join(query_conditions)}"

        query = f"""
                    SELECT type, count(*) as count
                    FROM messages_utc
                    WHERE type NOT IN ('new_chat_members', 'left_chat_member', 'new_chat_photo',
                                       'new_chat_title', 'migrate_from_group', 'pinned_message')
                          {query_where}
                    GROUP BY type
                    ORDER BY count DESC;
                 """

        with self.engine.connect() as con:
            df = pd.read_sql_query(query, con, params=sql_dict)
        df['Group Percent'] = df['count'] / df['count'].sum() * 100
        df.columns = ['type', 'Group Count', 'Group Percent']

        if user:
            sql_dict['user'] = user[0]
            query_conditions.append("from_user = %(user)s")

            query = f"""
                        SELECT type, count(*) as user_count
                        FROM messages_utc
                        WHERE type NOT IN ('new_chat_members', 'left_chat_member', 'new_chat_photo',
                                           'new_chat_title', 'migrate_from_group', 'pinned_message')
                              AND {' AND '.join(query_conditions)}
                        GROUP BY type
                        ORDER BY user_count DESC;
                     """
            with self.engine.connect() as con:
                df_u = pd.read_sql_query(query, con, params=sql_dict)
            df_u['User Percent'] = df_u['user_count'] / df_u['user_count'].sum() * 100
            df_u.columns = ['type', 'User Count', 'User Percent']

            df = df.merge(df_u, on="type", how="outer")

        a = list(zip(df.columns.values, ["Total"] + df.iloc[:, 1:].sum().to_list()))
        df = df.append({key: value for key, value in a}, ignore_index=True)
        df['Group Count'] = df['Group Count'].astype('Int64')
        try:
            df['User Count'] = df['User Count'].astype('Int64')
        except KeyError:
            pass
        text = df.to_string(index=False, header=True, float_format=lambda x: f"{x:.1f}")

        if user:
            return f"**Messages by type, {escape_markdown(user[1])} vs group:**\n```\n{text}\n```", None
        else:
            return f"**Messages by type:**\n```\n{text}\n```", None

    def get_random_message(self, lquery: str = None, start: str = None, end: str = None,
                           user: Tuple[int, str] = None, **kwargs) -> Tuple[str, None]:
        """
        Display a random message.
        :param lquery: Limit results to lexical query (&, |, !, <n>)
        :param start: Start timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        :param end: End timestamp (e.g. 2019, 2019-01, 2019-01-01, "2019-01-01 14:21")
        """
        query_conditions = []
        sql_dict = {}

        if lquery:
            query_conditions.append(f"text_index_col @@ to_tsquery('{lquery}')")

        if user:
            sql_dict['user'] = user[0]
            query_conditions.append("from_user = %(user)s")

        if start:
            sql_dict['start_dt'] = pd.to_datetime(start)
            query_conditions.append("date >= %(start_dt)s")

        if end:
            sql_dict['end_dt'] = pd.to_datetime(end)
            query_conditions.append("date < %(end_dt)s")

        query_where = ""
        if query_conditions:
            query_where = f"AND {' AND '.join(query_conditions)}"

        query = f"""
                SELECT date, from_user, text
                FROM messages_utc
                WHERE type = 'text'
                {query_where}
                ORDER BY RANDOM()
                LIMIT 1;
                """

        with self.engine.connect() as con:
            result = con.execute(query, sql_dict)
        try:
            date, from_user, text = result.fetchall()[0]
        except IndexError:
            return "No matching messages", None

        return f"*On {escape_markdown(date.strftime('%Y-%m-%d'))}, {escape_markdown(self.users[from_user][0])}" \
               f" gave these words of wisdom:*\n" \
               f"{escape_markdown(text)}\n",\
               None


def get_parser(runner: StatsRunner) -> InternalParser:
    parser = InternalParser(prog="/stats")
    parser.set_defaults(func=runner.get_chat_counts)
    subparsers = parser.add_subparsers(title="Statistics:")

    parser.add_argument('-v', '--version', action='version', version=__version__)

    for name, func in runner.allowed_methods.items():
        try:
            doc = inspect.getdoc(getattr(runner, func)).splitlines()
        except AttributeError:
            doc = None
        subparser = subparsers.add_parser(name, help=doc[0])
        subparser.set_defaults(func=getattr(runner, func))
        f_args = inspect.signature(getattr(runner, func)).parameters

        for _, arg in f_args.items():
            arg: inspect.Parameter
            if arg.name == 'self':
                continue
            if arg.name == 'user':
                group = subparser.add_mutually_exclusive_group()
                group.add_argument('-me', action='store_true', help='calculate stats for yourself')
                group.add_argument('-user', type=int, help=argparse.SUPPRESS)
            elif arg.name == 'autouser':
                subparser.set_defaults(me=True)
                subparser.add_argument('-user', type=int, help=argparse.SUPPRESS)
            elif arg.name == 'kwargs':
                pass
            else:
                arg_doc = None
                if doc:
                    for line in doc:
                        match = re.match(rf"^:param {arg.name}: (.*)", line)
                        if match:
                            arg_doc = match.group(1)

                if arg.annotation == bool:
                    subparser.add_argument(f"-{arg.name}".replace('_', '-'), action='store_true', help=arg_doc)
                else:
                    subparser.add_argument(f"-{arg.name}".replace('_', '-'), type=arg.annotation, help=arg_doc,
                                           default=arg.default)

    return parser



# !/usr/bin/env python
#
# A logging and statistics bot for Telegram based on python-telegram-bot.
# Copyright (C) 2020
# Michael DM Dryden <mk.dryden@utoronto.ca>
#
# This file is part of telegram-stats-bot.
#
# telegram-stats-bot is free software: you can redistribute it and/or modify
# it under the terms of the GNU Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Public License
# along with this program. If not, see [http://www.gnu.org/licenses/].
from copy import copy

import datetime
import logging
import json
import os

from sqlalchemy import MetaData, Table, Column, create_engine, BigInteger, TIMESTAMP, Text
from sqlalchemy_utils import database_exists
from sqlalchemy_utils.functions.orm import quote

from .parse import MessageDict
from .db import init_dbs

logger = logging.getLogger(__name__)
metadata = MetaData()

messages = Table('messages_utc', metadata,
                 Column('message_id', BigInteger),
                 Column('date', TIMESTAMP),
                 Column('from_user', BigInteger),
                 Column('forward_from_message_id', BigInteger),
                 Column('forward_from', BigInteger),
                 Column('forward_from_chat', BigInteger),
                 Column('caption', Text),
                 Column('text', Text),
                 Column('sticker_set_name', Text),
                 Column('new_chat_title', Text),
                 Column('reply_to_message', BigInteger),
                 Column('file_id', Text),
                 Column('type', Text))
user_events = Table('user_events', metadata,
                    Column('message_id', BigInteger),
                    Column('user_id', BigInteger),
                    Column('date', TIMESTAMP),
                    Column('event', Text))


def date_converter(o):
    if isinstance(o, datetime.datetime):
        return o.__str__()


class JSONStore(object):
    def __init__(self, path: str):
        self.store = path

    def append_data(self, name: str, data: MessageDict):
        with open(os.path.join(self.store, f"{name}.json"), 'a') as f:
            f.write(json.dumps(data, default=date_converter) + "\n")


class PostgresStore(object):
    def __init__(self, connection_url: str):
        self.engine = create_engine(connection_url, echo=False, isolation_level="AUTOCOMMIT")
        if not database_exists(self.engine.url):
            text = f"""
                    CREATE DATABASE {quote(self.engine, self.engine.url.database)}
                    ENCODING 'utf-8'
                    TEMPLATE {quote(self.engine, 'template1')}
                    """

            url = copy(self.engine.url)
            url.database = 'postgres'

            engine = create_engine(url, echo=False, isolation_level="AUTOCOMMIT")
            result_proxy = engine.execute(text)

            if result_proxy is not None:
                result_proxy.close()
            engine.dispose()

        init_dbs(self.engine)

    def append_data(self, name: str, data: MessageDict):
        data['date'] = str(data['date'])
        if name == 'messages':
            with self.engine.connect() as con:
                _ = con.execute(messages.insert(), data)
        elif name == 'user_events':
            with self.engine.connect() as con:
                _ = con.execute(user_events.insert(), data)
        else:
            logger.warning("Tried to append to invalid table %s", name)

    def update_data(self, name: str, data: MessageDict):
        data['date'] = str(data['date'])
        if name == 'messages':
            update_statement = messages.update()\
                                       .where(messages.c.message_id == data['message_id'])
            with self.engine.connect() as con:
                _ = con.execute(update_statement, data)
        elif name == 'user_events':
            update_statement = user_events.update()\
                                          .where(user_events.c.message_id == data['message_id'])
            with self.engine.connect() as con:
                _ = con.execute(update_statement, data)
        else:
            logger.warning("Tried to update to invalid table %s", name)




# !/usr/bin/env python
#
# A logging and statistics bot for Telegram based on python-telegram-bot.
# Copyright (C) 2020
# Michael DM Dryden <mk.dryden@utoronto.ca>
#
# This file is part of telegram-stats-bot.
#
# telegram-stats-bot is free software: you can redistribute it and/or modify
# it under the terms of the GNU Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Public License
# along with this program. If not, see [http://www.gnu.org/licenses/].

import sys

from typing import Tuple, Union, List
if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from datetime import datetime

import telegram


class MessageDict(TypedDict):
    message_id: int
    date: Union[str, datetime]
    from_user: int
    forward_from_message_id: Union[int, None]
    forward_from: Union[int, None]
    forward_from_chat: Union[int, None]
    caption: Union[str, None]
    text: Union[str, None]
    sticker_set_name: Union[str, None]
    new_chat_title: Union[str, None]
    reply_to_message: Union[int, None]
    file_id: Union[str, None]
    type: str


def parse_message(message: telegram.message.Message) -> Tuple[dict, List[dict]]:
    message_dict: MessageDict = {'message_id': message.message_id,
                                 'date': message.date,
                                 'from_user': None,
                                 'forward_from_message_id': message.forward_from_message_id,
                                 'forward_from': None,
                                 'forward_from_chat': None,
                                 'caption': message.caption,
                                 'text': message.text,
                                 'sticker_set_name': None,
                                 'new_chat_title': message.new_chat_title,
                                 'reply_to_message': None,
                                 'file_id': None,
                                 'type': None,
                                 }
    user_event_dict = [{}]

    if message.from_user:
        message_dict['from_user'] = message.from_user.id

    if message.forward_from:
        try:
            message_dict['forward_from'] = message.forward_from.id
        except AttributeError:
            pass
        try:
            message_dict['forward_from_chat'] = message.forward_from_chat.id
        except AttributeError:
            pass

    if message.reply_to_message:
        message_dict['reply_to_message'] = message.reply_to_message.message_id

    if message.text:
        message_dict['type'] = 'text'
    elif message.animation:
        message_dict['type'] = 'animation'
        message_dict['file_id'] = message.animation.file_id
    elif message.audio:
        message_dict['type'] = 'audio'
        message_dict['file_id'] = message.audio.file_id
    elif message.document:
        message_dict['type'] = 'document'
        message_dict['file_id'] = message.document.file_id
    elif message.game:
        message_dict['type'] = 'game'
    elif message.photo:
        message_dict['type'] = 'photo'
    elif message.sticker:
        message_dict['type'] = 'sticker'
        message_dict['file_id'] = message.sticker.file_id
        message_dict['sticker_set_name']: message.sticker.set_name
    elif message.video:
        message_dict['type'] = 'video'
    elif message.video_note:
        message_dict['type'] = 'video_note'
    elif message.voice:
        message_dict['type'] = 'voice'
    elif message.location:
        message_dict['type'] = 'location'
    elif message.poll:
        message_dict['type'] = 'poll'
    elif message.new_chat_title:
        message_dict['type'] = 'new_chat_title'
    elif message.new_chat_photo:
        message_dict['type'] = 'new_chat_photo'
    elif message.pinned_message:
        message_dict['type'] = 'pinned_message'
    elif message.new_chat_members:
        message_dict['type'] = 'new_chat_members'
        member: telegram.user.User
        user_event_dict = [{'message_id': message.message_id,
                            'user_id': u_id,
                            'date': message.date,
                            'event': 'joined'}
                           for u_id in [member.id for member in message.new_chat_members]]
    elif message.left_chat_member:
        message_dict['type'] = 'left_chat_member'
        user_event_dict = [{'message_id': message.message_id,
                            'user_id': message.left_chat_member.id,
                            'date': message.date,
                            'event': 'left'}]

    return message_dict, user_event_dict



# !/usr/bin/env python
#
# A logging and statistics bot for Telegram based on python-telegram-bot.
# Copyright (C) 2020
# Michael DM Dryden <mk.dryden@utoronto.ca>
#
# This file is part of telegram-stats-bot.
#
# telegram-stats-bot is free software: you can redistribute it and/or modify
# it under the terms of the GNU Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Public License
# along with this program. If not, see [http://www.gnu.org/licenses/].

import re

md_match = re.compile(r"(\[[^][]*]\(http[^()]*\))|([_*[\]()~>#+-=|{}.!\\])")


def escape_markdown(string: str) -> str:
    def url_match(match: re.Match):
        if match.group(1):
            return f'{match.group(1)}'
        return f'\\{match.group(2)}'

    return re.sub(md_match, url_match, string)



# !/usr/bin/env python
#
# A logging and statistics bot for Telegram based on python-telegram-bot.
# Copyright (C) 2020
# Michael DM Dryden <mk.dryden@utoronto.ca>
#
# This file is part of telegram-stats-bot.
#
# telegram-stats-bot is free software: you can redistribute it and/or modify
# it under the terms of the GNU Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser Public License for more details.
#
# You should have received a copy of the GNU Public License
# along with this program. If not, see [http://www.gnu.org/licenses/].

import logging

from sqlalchemy.engine import Engine


logger = logging.getLogger(__name__)


def init_dbs(engine: Engine):
    sql = """
        create table if not exists messages_utc
        (
            message_id              bigint,
            date                    timestamptz,
            from_user               bigint,
            forward_from_message_id bigint,
            forward_from            bigint,
            forward_from_chat       bigint,
            caption                 text,
            text                    text,
            sticker_set_name        text,
            new_chat_title          text,
            reply_to_message        bigint,
            file_id                 text,
            type                    text
        );
        
        alter table messages_utc
                add column if not exists text_index_col tsvector
                generated always as (to_tsvector('english', coalesce(text, ''))) stored;
        
        create index if not exists text_idx
            on messages_utc using gin (text_index_col);
        
        create index if not exists messages_utc_date_index
            on messages_utc (date);
        
        create index if not exists messages_utc_from_user_index
            on messages_utc (from_user);
        
        create index if not exists messages_utc_type_index
            on messages_utc (type);
            
        create table if not exists user_events
        (
            message_id bigint,
            user_id    bigint,
            date       timestamp with time zone,
            event      text
        );
        
        create index if not exists ix_user_events_message_id
            on user_events (message_id);
        
        create table if not exists user_names
        (
            user_id  bigint,
            date     timestamptz,
            username text,
            display_name text
        );
        
        create index if not exists user_names_user_id_date_index
            on user_names (user_id, date);
        
        """

    with engine.connect() as con:
        con.execute(sql)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['telegram_stats_bot']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy-Utils==0.37.6',
 'appdirs>=1.4.4,<2.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.4,<2.0.0',
 'pandas>=1.0.4,<2.0.0',
 'python-telegram-bot>=13.6,<14.0',
 'seaborn>=0.10.1,<0.11.0',
 'single-source>=0.2.0,<0.3.0',
 'sqlalchemy[postgresql_psycopg2binary]>=1.4,<2.0']

setup_kwargs = {
    'name': 'telegram-stats-bot',
    'version': '0.4.0',
    'description': 'A logging and statistics bot for Telegram based on python-telegram-bot.',
    'long_description': '==================\ntelegram-stats-bot\n==================\n\n.. image:: https://img.shields.io/pypi/v/telegram-stats-bot?style=flat-square\n   :target: https://pypi.org/project/telegram-stats-bot/\n   :alt: PyPi Package Version\n\n.. image:: https://img.shields.io/pypi/pyversions/telegram-stats-bot?style=flat-square\n   :target: https://pypi.org/project/telegram-stats-bot/\n   :alt: Supported Python versions\n\n.. image:: https://img.shields.io/github/commit-activity/m/mkdryden/telegram-stats-bot?style=flat-square\n   :target: https://github.com/mkdryden/telegram-stats-bot\n   :alt: GitHub Commit Activity\n\n.. image:: https://img.shields.io/pypi/l/telegram-stats-bot?style=flat-square\n   :target: https://www.gnu.org/licenses/gpl-3.0-or-later.html\n   :alt: LGPLv3 License\n\n\nTelegram-stats-bot is a simple bot that lives in your Telegram group, logging messages to a Postgresql database and\nserving statistical tables and plots to users as Telegram messages.\n\n.. image:: examples/chat_example.png\n    :alt: Bot conversation example\n\n-----------------\nTable of contents\n-----------------\n\n- `Introduction`_\n\n- `Basic Requirements`_\n\n- `Installation`_\n\n- `Setup`_\n\n- `Fetching Stats`_\n\n  - `counts`_\n\n  - `hours`_\n\n  - `days`_\n\n  - `week`_\n\n  - `history`_\n\n  - `titles`_\n\n  - `corr`_\n\n  - `delta`_\n\n  - `types`_\n\n  - `random`_\n\n- `License`_\n\n------------\nIntroduction\n------------\n\nThis software is intended to be run on a server, handling updates for a bot user with a single bot per channel\n(multi-channel support could be added at some point if there is interest), using the excellent\n`Python-telegram-bot <https://github.com/python-telegram-bot/python-telegram-bot>`_ library.\n\nThe bot is still in active development but at the moment, it features:\n\n- Message logging to Postgresql database with optional JSON file backup\n\n- Statistics output for users in group as Telegram messages, with optional filtering by date or limiting to the querying\n  user.\n  Some statistics are more useful than others, but they are mainly intended to be fun for users to play with.\n\n  - Tables:\n\n    - Most active users\n\n    - A user\'s message time correlation with other users\n\n    - A user\'s median message time difference with other users\n\n  - Plots:\n\n    - Message activity by hour of day\n\n    - Message activity by day of week\n\n    - Message activity over the week by hour and day\n\n    - Message activity history\n\n------------------\nBasic Requirements\n------------------\n\n- Python 3.7+\n\n- A Telegram bot token with privacy mode disabled (needed to log messages)\n\n  - See `here <https://core.telegram.org/bots#6-botfather>`_ for details\n\n- Postgresql (Tested with 12.3, but there shouldn\'t be anything that won\'t work with 9.4 or up)\n\n  - This can be on a different system than telegram-stats-bot and requires either table creation permissions on a\n    database or database can be pre-initialized following the setup in ``db.py``\n\n------------\nInstallation\n------------\n\nThe easiest way to install or upgrade is with pip:\n\n.. code:: shell\n\n    $ pip install telegram-stats-bot --upgrade\n\nOr you can install the latest git version using `poetry <https://python-poetry.org/>`_ (installed to Python 3.7 or later):\n\n.. code:: shell\n\n    $ git clone https://github.com/mkdryden/telegram-stats-bot.git\n    $ poetry install\n    \nInstalling directly with ``setup.py`` should also work from the Pypi sdist (but why?).\n\n-----\nSetup\n-----\nOnce installed, you can run the bot by calling the ``main`` module with a few required arguments:\n\n.. code:: shell\n\n    $ python -m telegram_stats_bot.main BOT_TOKEN CHAT_ID POSTGRESQL_URL\n\n- BOT_TOKEN: Your bot\'s token e.g., ``110201543:AAHdqTcvCH1vGWJxfSeofSAs0K5PALDsaw``\n\n- CHAT_ID: The chat id to monitor (will be a large integer, possibly negative, if unknown, set to 0 and see below)\n\n- POSTGRESQL_URL: Connection information in the form: ``postgresql://USERNAME:PASSWORD@ADDRESS/DB_NAME``\n\n  - if DB_NAME exists, there must not be tables called ``messages_utc``, ``user_events``, or ``user_names``\n    with incorrect columns\n\nOn startup, the bot will attempt to create the database and tables, if they do not already exist.\nIf you do not know the chat\'s id and have set it to 0 as mentioned above, you can send the ``/chatid`` command inside\nthe group, and the bot will reply with it, then restart the bot with the id.\nIf you have forgotten to disable privacy mode, an error will be logged in the terminal.\n\nThe bot will now log all messages in the group, but will only respond to users who have sent a message that has been\nlogged previously (and this list is only updated once an hour, so if you\'re impatient, you can restart the bot after\nyou\'ve sent a message to trigger the update).\nYou can see if messages are being logged correctly by reviewing the terminal output.\nYou should see a line like ``2020-06-04 02:08:39,212 - __main__ - INFO - 8``, whenever a message is logged.\n\n--------------\nFetching Stats\n--------------\nTo fetch stats, simply message the bot, either inside the group being logged, or in a direct message, using the\n``/stats`` command.\n``/stats`` with no arguments prints the table of most active users and other statistics are available through various\nsubcommands.\nAll commands are documented and the built in help can be displayed with ``/stats -h`` or ``stats <subcommand> -h``.\n\nMost commands have optional arguments that change the behaviour of the output.\nNearly all have:\n\n- ``-start`` and ``-end`` followed by a timestamp (e.g., 2019, 2019-01, 2019-01-01, "2019-01-01 14:21") specify the\n  range of data to fetch, otherwise all available data will be used.\n  Either or both options can be given.\n\n- ``lsquery`` followed by a lexical query (using Postgres\'\n  `tsquery syntax <https://www.postgresql.org/docs/12/datatype-textsearch.html#DATATYPE-TSQUERY>`_)\n  limits results to matching messages.\n\n- ``-me`` calculates statistics for the user sending the command, rather than all chat users.\n\nSample outputs of each available subcommand follow.\n\ncounts\n------\n``/stats counts`` returns a list of the most active users in the group.\n\n.. code::\n\n                User  Total Messages  Percent\n          @ACoolUser           42150      7.0\n          @NumberOne           37370      6.2\n   @WinstonChurchill           32668      5.4\n            @AAAAAAA           32134      5.4\n  @WhereAreMyManners           30481      5.1\n  @TheWorstOfTheBest           28705      4.8\n\nhours\n-----\n``/stats hours`` returns a plot of message frequency for the hours of the day.\n\n.. image:: examples/hours.png\n   :alt: Example of hours plot\n\ndays\n----\n``/stats days`` returns a plot of message frequency for the days of the week.\n\n.. image:: examples/days.png\n   :alt: Example of days plot\n\nweek\n----\n``/stats week`` returns a plot of total messages over the data period by day of week and hour of day.\n\n.. image:: examples/week.png\n   :alt: Example of week plot\n\nhistory\n-------\n``/stats history`` returns a plot of messages versus date.\n\n.. image:: examples/history.png\n   :alt: Example of history plot\n\ntitles\n------\n``/stats titles`` returns a plot of group titles over time.\n\n.. image:: examples/titles.png\n   :alt: Example of title history plot\n\ncorr\n----\n``/stats corr`` returns a list of users with the highest and lowest message time correlations with the requesting user.\n\n.. code::\n\n    User Correlations for @TheManWhoWasThursday\n    HIGHEST CORRELATION:\n    @MyGoodFriend         0.335\n    @Rawr                 0.302\n    @MangesUnePoutine     0.284\n    @GreenBlood           0.251\n    @TooMuchVacuum        0.235\n\n    LOWEST CORRELATION:\n    @Shiny                0.146\n    @BlueDog              0.142\n    @CoolCat              0.122\n    @EatMe                0.116\n    @JustPassingBy        0.106\n\ndelta\n-----\n``/stats delta`` returns a list of users with the shortest differences in message times with the requesting user.\n\n.. code::\n\n    Median message delays for @KingLeer and:\n    @PolyamorousPasta     00:03:23\n    @AggressiveArgon      00:04:43\n    @AdjectiveNoun        00:08:27\n    @SuperSalad           00:09:05\n    @ABoredProgrammer     00:09:06\n\ntypes\n-----\n``/stats types`` returns a table of messages by type, comparing the requesting user with the full group.\n\n.. code::\n\n Messages by type, @AUser vs group:\n       type  Group Count  Group Percent  User Count  User Percent\n       text     528813.0           88.3     13929.0          83.4\n    sticker      34621.0            5.8      1226.0           7.3\n      photo      25995.0            4.3      1208.0           7.2\n  animation       6983.0            1.2       274.0           1.6\n      video       1325.0            0.2        48.0           0.3\n      voice        475.0            0.1         2.0           0.0\n   location        252.0            0.0         2.0           0.0\n video_note         84.0            0.0         1.0           0.0\n      audio         62.0            0.0         1.0           0.0\n       poll         29.0            0.0         1.0           0.0\n   document          1.0            0.0         1.0           0.0\n      Total     598640.0          100.0     16693.0         100.0\n\nrandom\n------\n``/stats random`` prints a random message from the database.\n\n----------\nThe Future\n----------\nTelegram-stats-bot is a work in progress.\nNew stats will be added, but no guarantees that the database structure will stay constant if Telegram\'s message\nstructure changes or I need to change something to make a new statistic work.\n\n=======\nLicense\n=======\nTelegram-stats-bot is free software:\nYou can redistribute it and/or modify it under the terms of the\n`GNU General Public License v3.0 or later <https://www.gnu.org/licenses/gpl-3.0.html>`_.\nDerivative works must also be redistributed under the GPL v3 or later.\n',
    'author': 'Michael DM Dryden',
    'author_email': 'mk.dryden@utoronto.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mkdryden/telegram-stats-bot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
